using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;
using System.Windows.Forms; 
using System.ComponentModel;
using System.IO;

namespace Sequence
{
    public class DataClient
    {
        /// <summary>
        /// Connection with the system
        /// </summary>
        private Socket socket;

        /// <summary>
        /// The ip of the system
        /// </summary>
        private String systemIp;

        /// <summary>
        /// The port that we listen to
        /// </summary>
        private int systemPort;

        /// <summary>
        /// The size of one packet
        /// </summary>
        private int size;
        private int totalSize;

        /// <summary>
        /// Background worker that enables to handle the socket reception
        /// </summary>
        private BackgroundWorker worker;

        /// <summary>
        /// Number of received data
        /// </summary>
        public int recepted;

        /// <summary>
        /// Number of received data
        /// </summary>
        public int receptedData;
        public int totalReceptedData;

        /// <summary>
        /// Binary writer
        /// </summary>
        private BinaryWriter _writer;
        
        /// <summary>
        /// Store the output path
        /// </summary>
        public String outputpath;
        
        /// <summary>
        /// Delegate to become event when a packet is received
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void PacketReceivedEventHandler(object sender, DataClientEventArgs e);

        /// <summary>
        /// Definition of the evenement
        /// </summary>
        public event PacketReceivedEventHandler packetReceived;

        /// <summary>
        /// Open a database
        /// </summary>
        /// <returns></returns>
        public int OpenFileToWrite()
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Filter = "csv files (*csv)|*.csv|All files (*.*)|*.*";

            if (dlg.ShowDialog() != DialogResult.OK)
                return -1;
            
            this.outputpath = dlg.FileName;

            return 0; 
        }

        /// <summary>
        /// Constructor of the class
        /// </summary>
        /// <param name="ipAddress">The ip address of the system</param>
        /// <param name="port">The port used</param>
        /// <param name="packetSize">The size of one packet to receive in byte</param>
        public DataClient(String ipAddress, int port, int packetSize, int totalSize, BinaryWriter writer)
        {
            socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            // Copy params
            this.systemIp          = ipAddress;
            this.systemPort        = port;
            this.size              = packetSize;
            this.totalSize         = totalSize;
            this._writer           = writer;
            this.receptedData      = 0;
            this.totalReceptedData = 0;

            // Create the background worker
            worker = new BackgroundWorker();
            worker.WorkerSupportsCancellation = true;
            worker.WorkerReportsProgress = true;
            worker.DoWork += new DoWorkEventHandler(worker_DoWork);
        }

        /// <summary>
        /// Work of the background worker
        /// Receive data from the socket
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            // Allocate space
            byte[] buffer = new byte[2097152];

            DataClientEventArgs eventArgs = new DataClientEventArgs();
            eventArgs.data = new byte[this.size];
            int eventBufferFreeSize = this.size;
            int eventBufferIndex = 0;

            try
            {
                while (this.worker.CancellationPending == false)
                {
                    int offset = 0;

                    // Receive a large chunk
                    // Receiving a large chunk to then convert it is faster than receiving small chunk of data.
                    recepted = socket.Receive(buffer);
                    
                    // Write in file if needed
                    if (this._writer != null)
                        this._writer.Write(buffer, 0, recepted);

                    while (recepted > 0)
                    {                        
                        int chunk = (recepted > eventBufferFreeSize) ? eventBufferFreeSize : recepted;

                        // Copy buffer
                        Buffer.BlockCopy(buffer, offset, eventArgs.data, eventBufferIndex, chunk);
                        
                        // Update
                        recepted            -= chunk;
                        offset              += chunk;
                        eventBufferIndex    += chunk;
                        eventBufferFreeSize -= chunk;

                        // Get number of data acquired total and for this buffer
                        this.receptedData      += chunk;
                        this.totalReceptedData += chunk;

                        // Check end of transfer
                        if ((this.totalReceptedData == this.totalSize))
                            eventBufferFreeSize = 0;

                        // Do we have to send ?
                        if (eventBufferFreeSize == 0)
                        {
                            // Put data count
                            eventArgs.dataCount = this.receptedData;
                            // Generate event
                            OnPacketReceived(eventArgs);

                            // Reset values
                            eventBufferFreeSize = this.size;
                            this.receptedData   = 0;
                            eventBufferIndex    = 0;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error : " + ex.Message);
            }
        }

        /// <summary>
        /// Signal the event to the listeners
        /// </summary>
        /// <param name="e"></param>
        public virtual void OnPacketReceived(DataClientEventArgs e)
        {
            if (packetReceived != null)
            {
                packetReceived(this, e);
            }
        }

        /// <summary>
        /// Start to listen from the packet
        /// </summary>
        /// <returns></returns>
        public int StartListening()
        {
            // Try to connect
            try
            {
                socket.Connect(systemIp, systemPort);
            }
            catch (Exception ex)
            {
                Console.WriteLine("socket connection exception" + ex.Message);
                return -1;
            }

            worker.RunWorkerAsync();
            return 0;
        }

        /// <summary>
        /// Close connection
        /// </summary>
        public void Close()
        {
            // Close the connection
            try
            {
                socket.Close();
            }
            catch (Exception) { };

            // Stop the worker
            worker.CancelAsync();
            worker.Dispose();

            // Reset variables
            this.receptedData      = 0;
            this.totalReceptedData = 0;
        }

        /// <summary>
        /// Print available bytes in the socket
        /// </summary>
        public void PrintAvailable()
        {
            try
            {
                Console.WriteLine("Available : " + socket.Available);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Exception : " + ex.Message);
            }
        }
    }
}
