#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <MSXE351x.nsmap> // this file must be included only once in the project
#include <assert.h>
#include <sys/stat.h>
#include <signal.h>
#include "utils.h"

/* please see utils.h */
int DUMMY = 0;

/* IP Address of the module */
char * ipAddress = "192.168.99.99";

int watchdogInitialized = 0;

//-------------------------------------------------------------------------------------------------------

/** Initialize and start the watchdog.
 *
 * @param[in] soapEnv The current SOAP environment
 * @param[in] endPoint : SOAP endpoint
 *
 * @retval 0:
 */
int sampleWatchdogInitialization(struct soap *soapEnv, char *endPoint)
{
    struct MSXE351x__Response response;
    unsigned long timeBase = 0;
    unsigned long timeValue = 0;

    printf("Watchdog configuration\n");

    do
    {
        printf("Time base of the watchdog delay (0 for us, 1 for ms, 2 for s): ");
        scanf("%lu", &timeBase);
        printf("\n");
    } while (timeBase > 2);

    do
    {
        printf("Time base value of the watchdog delay (0 to 65535): ");
        scanf("%lu", &timeValue);
        printf("\n");
    } while (timeValue > 0xFFFF);
    assert_gsoap(soapEnv,
            soap_call_MSXE351x__IOWatchdogInitAndStart(soapEnv,
                    endPoint,
                    NULL,
                    timeBase,
                    timeValue,
                    0,
                    0,
                    &response));
    assert_soap_command(soapEnv, (struct DefaultResponse *)&response);
    printf("soap_call_MSXE351x__IOWatchdogInitAndStart OK\n");
    watchdogInitialized = 1;
    return 0;
}

//------------------------------------------------------------------------------------------
/** writeOneFunction : write value on only one channel to the analog output of the MSX-E351x
 *
 * @param[in] soapEnv The current SOAP environment
 * @param[in] endPoint : SOAP endpoint
 * @param[in] nbrOfChannel : Number of channel on the board
 *
 * @return 0
 *
 */
int writeOneFunction(struct soap *soapEnv, char *endPoint, unsigned long nbrOfChannel)
{
    struct MSXE351x__Response response;
    unsigned long channel = 0;
    unsigned long outputType = 0;
    unsigned long polarity = 0;
    unsigned long triggerMask = 0;
    unsigned long triggerEdgeSelection = 0;
    unsigned long triggerCount = 0;
    unsigned long channelValue = 0x7FFF;
    response.iReturnValue = 0;

    printf("\n");
    assert_gsoap(soapEnv,
            soap_call_MSXE351x__AnalogOutputWrite1Value(soapEnv,
                    endPoint,
                    NULL,
                    channel,
                    outputType,
                    polarity,
                    triggerMask,
                    triggerEdgeSelection,
                    triggerCount,
                    channelValue,
                    0,
                    0,
                    0,
                    &response));
    assert_soap_command(soapEnv, (struct DefaultResponse *)&response);
    printf("Write %lu on channel %lu\n", channelValue, channel);
    return 0;
}

//-------------------------------------------------------------------------------------------------------

/** Get watchdog status and value.
 *
 * @param[in] soapEnv The current SOAP environment
 * @param[in] endPoint : SOAP endpoint
 *
 * @retval 0:	No error.
 * @retval 1:	Error.
 */
int sampleGetWatchdogStatusAndValue(struct soap *soapEnv, char *endPoint)
{
    struct MSXE351x__IOWatchdogGetStatusAndValueResponse WatchdogStatusAndValueResponse;
    WatchdogStatusAndValueResponse.sResponse.iReturnValue = 0;

    assert_gsoap(soapEnv, soap_call_MSXE351x__IOWatchdogGetStatusAndValue(soapEnv, endPoint, NULL, 0, &WatchdogStatusAndValueResponse));
    assert_soap_command(soapEnv, (struct DefaultResponse*) &(WatchdogStatusAndValueResponse.sResponse));
    printf("\rWatchdog status %lu value = %5lu info = %lu",
           WatchdogStatusAndValueResponse.ulStatus,
           WatchdogStatusAndValueResponse.ulValue,
           WatchdogStatusAndValueResponse.ulInfo);

    return ((WatchdogStatusAndValueResponse.ulStatus >> 1) & (WatchdogStatusAndValueResponse.ulStatus >> 0) & 1);
}

//-------------------------------------------------------------------------------------------------------

/** Stop and release the watchdog.
 *
 * @param[in] soapEnv The current SOAP environment
 * @param[in] endPoint : SOAP endpoint
 *
 * @retval 0:	No error.
 * @retval 1:	Error.
 *
 */
int sampleStopAndReleaseWatchdog(struct soap *soapEnv, char *endPoint)
{
    struct MSXE351x__Response response;
    printf("\n");

    assert_gsoap(soapEnv, soap_call_MSXE351x__IOWatchdogStopAndRelease(soapEnv, endPoint, NULL, 0, &response));
    assert_soap_command(soapEnv, (struct DefaultResponse*) &response);
    printf("soap_call_MSXE351x__IOWatchdogStopAndRelease OK\n");

    return 0;
}

//-----------------------------------------------------------------------------------------
/** signal handler : stops the current acquisition and cleanup the SOAP connection.
 *
 */
static void do_signal(int sig)
{
    struct soap soapEnv;
    mx_soap_init(&soapEnv);
    char endPoint[30];
    struct MSXE351x__Response response;

    printf("signal received:%s\n", strsignal(sig));
    printf("stopping current acquisition\n");

    /* generates the endpoint string
     * SOAP "endpoints" are of the form NAME:PORT -
     * PORT is in the case of the SOAP command server: 5555
     */
    if (snprintf(endPoint, sizeof(endPoint), "%s:5555", ipAddress) < 0)
    {
        fprintf(stderr, "error: %s:%d!\n", __FUNCTION__, __LINE__);
        exit(1);
    }
    /* note: the soapenv variable should have been correctly initialised before the signal handler was set */

    if (watchdogInitialized == 1)
    {
        /* Stop and release the watchdog */
        assert_gsoap(&soapEnv, soap_call_MSXE351x__IOWatchdogStopAndRelease(&soapEnv, endPoint, NULL, 0, &response));
        assert_soap_command(&soapEnv, (struct DefaultResponse*) &response);
        printf("soap_call_MSXE351x__IOWatchdogStopAndRelease OK\n");
    }
    mx_soap_clean(&soapEnv);

    exit(1);
}

//-----------------------------------------------------------------------------------------
/** For all signals set the handler do_signal.
 *
 * This allows to handle CTRL+C nicely
 *
 * @note This function must be called after the first call to a function of the soap_call() familly.
 */
static void setup_signal(void)
{
    struct sigaction action;
    /* catch all signal */
    sigfillset(&action.sa_mask); /* all other signal will be blocked when handling the one received */
    action.sa_handler = do_signal;
    action.sa_flags = 0;
    {
        int i;
        for (i = 1; i < NSIG; i++)
        {
            if (SIGKILL == i) /* this signal can not be blocked */
                continue;
            if (SIGSTOP == i) /* this signal can not be blocked */
                continue;
            sigaction(i, &action, NULL); /* TODO: manage error */
        }
    }
}

int main(int argc, char **argv)
{
    char endPoint[30];
    struct soap soapEnv;
    struct MXCommon__ByteArrayResponse response;
    int nbrOfChannel = 8;
    int RunDown = 0;
    /* Print sample information */
    printf("+---------------------------------------------------+\n");
    printf("| MSX-E351x                                         |\n");
    printf("| WatchDog sample					                |\n");
    printf("+---------------------------------------------------+\n");
    printf("\n");

    if (argc != 2)
    {
        printf("\nThe command should be like this : ./WriteOne XXX.XXX.XXX.XXX\n\n");
        exit(1);
    }

    memset(endPoint, 0, sizeof(endPoint));

    /* this structure the current SOAP connection */
    mx_soap_init(&soapEnv);

    /* Get the IP address */
    ipAddress = strdup(argv[1]);

    if (snprintf(endPoint, sizeof(endPoint), "%s:5555", ipAddress) < 0)
    {
        fprintf(stderr, "error: %s:%d!\n", __FUNCTION__, __LINE__);
        exit(1);
    }

    /* ask the module for its type */

    assert_gsoap(&soapEnv, soap_call_MXCommon__GetModuleType(&soapEnv, endPoint, NULL, &DUMMY,&response) );
    printf("Returned string: %s\n", response.sArray.__ptr);

    /* who is currently connected ? */
    assert_gsoap(&soapEnv,soap_call_MXCommon__GetClientConnections( &soapEnv,NULL, NULL, &DUMMY,&response ));
    assert_soap_command(&soapEnv,&(response.sResponse) );
    printf("Current connections: %s\n\n", response.sArray.__ptr);

    /* Initialize and start the watchdog */
    setup_signal();
    if (sampleWatchdogInitialization(&soapEnv, endPoint) == 0)
    {
        printf("Press a key to start\n");
        printf("Once started, press any key to set the first output to 10V (trigger the watchdog) or ESC to quit\n");
        getch();

        for (;;)
        {
            RunDown = sampleGetWatchdogStatusAndValue(&soapEnv, endPoint);
            /* Press any key to set all output to 1 */
            if (kbhit())
            {
                if (getch() == 27)
                    break;
                if (RunDown == 0)
                    writeOneFunction(&soapEnv, endPoint, nbrOfChannel);
            }

            /* Release the memory allocated by the gSOAP calls */
            soap_destroy(&soapEnv);
            soap_end(&soapEnv);
        }

        /* Stop and release the watchdog */
        sampleStopAndReleaseWatchdog(&soapEnv, endPoint);
    }
    mx_soap_clean(&soapEnv);

    /* Print End programme informations */
    printf("\n\n");
    printf("+---------------------------------------------------+\n");
    printf("| Program ended                                     |\n");
    printf("+---------------------------------------------------+\n");

    return 0;
}
